#!/usr/bin/perl
#
# Implements commands to be used via ${readsock
#
# Copyright 2003 Alun Jones, auj@aber.ac.uk
#
# This may be freely copied and modified.
#
# The problem:
#
# Calling ${perl leaves you with possible concurrency problems. If
# you are opening a DBM file, you need to lock it first. Under heavy
# load, or if the DBM file is corrupt, you end up with a single wedged
# exim process holding a lock and able to block all other deliveries.
# Or, if the DBM file is corrupted enough to crash the running process,
# you end up with all delivery attempts crashing.
#
# With exim 4 comes readsocket. This chucks data into a unix domain
# socket and reads the answer back. If no answer comes within a specified
# timeout, then it lets you default somewhere. So...
# replace ${perl{ with ${readsocket{ (with sensible timeout and failure
# actions). If this daemon dies or freezes or whatever, ${readsocket will 
# timeout and let mail carry on going.
# 

use lib "/usr/local/bin";
use Greylist;
use Socket;
use IO::Handle;
use POSIX ':signal_h';
use IPC::SysV qw(IPC_PRIVATE IPC_RMID S_IRWXU);

my $sockname = "/var/run/greylistd.sock";
my $pidfile = "/var/run/greylistd.pid";

my %cmds = (
  "GREYLIST" => \&Greylist::defercheck,
  "PREGREYLIST" => \&Greylist::pregreylist,
  "PING" => \&ping,
);

if($ARGV[0] eq '--help' || ($ARGV[0] ne '--cleanup' && $ARGV[0] ne '--forget' && $ARGV[0] ne '--daemon'))
  {
    print "Usage: greylistd.pl [OPTIONS]\n";
    print "--help -h        Hilfe\n";
    print "--daemon         Startet den Daemon\n";
    print "--cleanup        Startet den Cleanup fr die IP in Greylist\n";
    print "--forget <ip>    Lscht IP von Greylist\n\n";
    exit;
  }

if($ARGV[0] eq '--daemon')
{
  $SIG{'INT'}  = \&sigQUIT_handler;
  $SIG{'QUIT'} = \&sigQUIT_handler;
  $SIG{'TERM'} = \&sigQUIT_handler;
  
  if (open(F, "<$pidfile"))
  {
      my $running = <F>;
      close(F);
      chop($running);
      die "Another greylistd is running (pid = $running)\n" if (-e "/proc/$running");
  }

  daemonize();
  socket(UNIX, PF_UNIX, SOCK_STREAM, 0) || die "socket: $!";
  unlink($sockname);
  bind(UNIX, sockaddr_un($sockname)) || die "bind: $!";
  chmod(0666, $sockname);
  listen(UNIX, SOMAXCONN) || die "listen: $!";
  
  # Record out PID.
  open(F, ">$pidfile") || die "Can't write $pidfile: $!\n";
  print F "$$\n";
  close(F);
  
  close(STDIN);
  close(STDOUT);
  close(STDERR);
}

$SIG{'PIPE'} = sub { };
$SIG{'CHLD'} = 'IGNORE';

if($ARGV[0] eq '--cleanup')
{
  print "Cleanup...\n";
  Greylist::cleanup();
}
if($ARGV[0] eq '--forget')
{
  print "Forgetting IP $ARGV[1]...\n";
  Greylist::forget($ARGV[1]);
}

if($ARGV[0] eq '--daemon')
{
  while (1)
  {
      accept(C, UNIX);
      $pidCh = fork();
      if($pidCh == 0)
      {
        sysread(C, $_, 1024);
  
        s/^\s*//;
        my ($cmd, @args) = split('\s', $_);
        if (defined($cmds{$cmd}))
        {
          my $result = &{$cmds{$cmd}}(@args);
          syswrite(C, $result);
          close(C);
        }
        exit(0);
      }
      else
      {
        close(C);
      }
  }
}

sub daemonize
{
  chdir '/'               or die "Can't chdir to /: $!";
  open STDIN, '/dev/null' or die "Can't read /dev/null: $!";
  open STDOUT, '>/dev/null'
                                       or die "Can't write to /dev/null: $!";
  defined(my $pid = fork) or die "Can't fork: $!";
  exit if $pid;
  setsid                  or die "Can't start a new session: $!";
  open STDERR, '>&STDOUT' or die "Can't dup stdout: $!";
}

sub sigQUIT_handler
{
  close(UNIX);
  close(WRITE);
  unlink($sockname);
  unlink($pidfile);
  exit(0);
}

sub ping
{
  return 'PONG';
}
